#include <fstream>
#include "ground.h"


BGround::BGround()
{
	grid_size=20;
	clear();
}


void BGround::draw() const
{
}

void BGround::draw(const ist::i3d::PerspectiveCamera& camera) const
{
}


bool BGround::load(const std::string& fname) {
	std::ifstream f(fname.c_str(), std::ios::binary);
	if(!f)
		return false;

	int version=0;
	f.read((char*)&version, sizeof(int));
	if(version!=101)
		return false;

	ist::vector2 t_mn, t_mx;
	float gs;
	f.read((char*)t_mn.v, sizeof(float)*2);
	f.read((char*)t_mx.v, sizeof(float)*2);
	f.read((char*)&gs, sizeof(float));
	resize(t_mn, t_mx, gs);

	for(size_t i=0; i<vertex.size(); ++i) {
		f.read((char*)&vertex[i].pos.y, sizeof(float));
		f.read((char*)vertex[i].tex.v, sizeof(float)*2);
		f.read((char*)&vertex[i].attr, sizeof(int));
	}
	setNormal();
	return true;
}


void BGround::clear() {
	x_div = 0;
	z_div = 0;
	vertex.clear();
	index.clear();
}

void BGround::init(const ist::vector2& new_min, const ist::vector2& new_max, float gs) {
	min_edge = new_min;
	max_edge = new_max;
	grid_size = gs;
	x_div = std::max(int((max_edge.x-min_edge.x)/grid_size), 0)+1;
	z_div = std::max(int((max_edge.y-min_edge.y)/grid_size), 0)+1;
	vertex.clear();
	vertex.resize(x_div*z_div);

	initVertex();
	setIndex();
}

void BGround::resize(const ist::vector2& new_min, const ist::vector2& new_max, float gs) {
	BGround old=*this;
	init(new_min, new_max, gs);

	for(int i=0; i<z_div; ++i) {
		for(int j=0; j<x_div; ++j) {
			ist::vector3& v=(*this)[i][j].pos;
			v.y = old.getHeight(ist::vector2(v.x, v.z));
		}
	}
	setNormal();
}


BGround::coord_t BGround::getCoord(const ist::vector2& pos) const
{
	ist::vector2 rel_pos = pos-min_edge;
	return coord_t(int(rel_pos.x/grid_size), int(rel_pos.y/grid_size));
}

float BGround::getHeight(const ist::vector4& pos) const
{
	return getHeight(ist::vector2(pos.x, pos.z));
}

float BGround::getHeight(const ist::vector2& _target_pos) const
{
	ist::vector2 target_pos = _target_pos;
	coord_t coord = getCoord(target_pos);
	if(coord.x<0 || coord.z<0
	|| coord.x>x_div-2 || coord.z>z_div-2)
		return 0.0f;

	float h1 = (*this)[coord.z][coord.x].pos.y;
	float h2 = (*this)[coord.z][coord.x+1].pos.y;
	float h3 = (*this)[coord.z+1][coord.x].pos.y;
	float h4 = (*this)[coord.z+1][coord.x+1].pos.y;

	target_pos-=min_edge;
	ist::vector2 near_grid(grid_size*coord.x, grid_size*coord.z);
	float h5 = h1 + (h2-h1)/grid_size * (target_pos.x-near_grid.x);
	float h6 = h3 + (h4-h3)/grid_size * (target_pos.x-near_grid.x);
	return h5 + (h6-h5)/grid_size * (target_pos.y-near_grid.y);
}

ist::vector4 BGround::getRayCross(const ist::vector4 &eye, const ist::vector4 &ray) const
{
	ist::vector4 c_pos = eye;
	for(int i=0; i<50; ++i) {
		c_pos+=ray*50.0f;
		if(c_pos.y<getHeight(c_pos)) {
			for(int j=0; i<50; ++j) {
				c_pos-=ray;
				if(c_pos.y>getHeight(c_pos)) {
					return c_pos;
				}
			}
		}
	}
	return ist::vector4(0, -10000.0f, 0);
}


const ist::vector2& BGround::getMin() const
{
	return min_edge;
}

const ist::vector2& BGround::getMax() const
{
	static ist::vector2 t_mx;
	t_mx = getMin()+ist::vector2((x_div-1)*grid_size, (z_div-1)*grid_size);
	return t_mx;
}

void BGround::getHeightRange(float &low, float &high) const
{
	for(size_t i=0; i<vertex.size(); ++i) {
		float h = vertex[i].pos.y;
		if(i==0)
			low=high=h;
		if(h>high)
			high=h;
		if(h<low)
			low=h;
	}
}




void BGround::setNormal()
{
	setNormal(coord_t(0, 0), coord_t(getXDiv(), getZDiv()));
}

void BGround::setNormal(const coord_t& mn, const coord_t& mx)
{
	for(int i=mn.z; i<mx.z; ++i) {
		for(int j=mn.x; j<mx.x; ++j) {
			ist::vector3 c1;
			ist::vector3 c2;
			if(i!=0 && i!=z_div-1 && j!=0 && j!=x_div-1) {
				c1 = (*this)[i][j+1].pos - (*this)[i][j-1].pos;
				c2 = (*this)[i+1][j].pos - (*this)[i-1][j].pos;
				(*this)[i][j].nml = ist::vector3().cross(c2, c1).normalize();
			}
			else {
				(*this)[i][j].nml = ist::vector3(0.0f, 1.0f, 0.0f);
			}
		}
	}
}

void BGround::setIndex()
{
	index.resize(std::max((x_div-1)*(z_div-1)*4, 0));
	for(int i=0; i<z_div-1; ++i) {
		for(int j=0; j<x_div-1; ++j) {
			int bi=(x_div-1)*i*4+j*4;
			index[bi+0]=x_div*i + j;
			index[bi+1]=x_div*(i+1) + j;
			index[bi+2]=x_div*(i+1) + j+1;
			index[bi+3]=x_div*i + j+1;
		}
	}
}

void BGround::initVertex() {
	for(int i=0; i<z_div; ++i) {
		for(int j=0; j<x_div; ++j) {
			vertex_t &v = (*this)[i][j];
			v.pos = ist::vector3(min_edge.x+grid_size*j, 0, min_edge.y+grid_size*i);
			v.tex.x = 0.01f;
			v.tex.y = 0.01f;
			v.nml.y = 1.0f;
		}
	}
}

