
#include <string>
#include "ist/bitmap.h"
#include "ist/ist_math.h"
#include "ist/i3d_camera.h"



class IGround
{
public:
	virtual ~IGround() {}

	virtual bool load(const std::string& filename)=0;
	virtual void draw() const=0;
	virtual void draw(const ist::i3d::PerspectiveCamera& camera) const=0;

	virtual float getHeight(const ist::vector2& pos) const=0;
	virtual float getHeight(const ist::vector4& pos) const=0;
	virtual const ist::vector2& getMin() const=0;
	virtual const ist::vector2& getMax() const=0;
	virtual int getXDiv() const=0;
	virtual int getZDiv() const=0;
	virtual float getGridSize() const=0;

};



class BGround : public IGround
{
public:
	struct vertex_t
	{
		vertex_t() : attr(0) {}

		ist::vector3 pos;
		ist::vector3 nml;
		ist::vector2 tex;
		int attr;
	};

	struct coord_t
	{
		coord_t(int _x=0, int _z=0) : x(_x), z(_z) {}
		int x, z;
	};


	class Block
	{
	public:
		Block(BGround& g, coord_t mn, coord_t mx) : ground(&g), _min(mn), _max(mx) {}

	private:
		BGround *ground;
		coord_t _min, _max;
	};

	typedef std::vector<vertex_t> vertex_cont;
	typedef std::vector<int> index_cont;


	bool load(const std::string& fname);
	void draw() const;
	void draw(const ist::i3d::PerspectiveCamera& camera) const;
	float getHeight(const ist::vector4& pos) const;
	float getHeight(const ist::vector2& _target_pos) const;
	const ist::vector2& getMin() const;
	const ist::vector2& getMax() const;
	int getXDiv() const { return x_div; }
	int getZDiv() const { return z_div; }
	float getGridSize() const { return grid_size; }

	BGround();
	void clear();
	void init(const ist::vector2& new_min, const ist::vector2& new_max, float gs);
	void resize(const ist::vector2& new_min, const ist::vector2& new_max, float gs);
	void setNormal();
	void setNormal(const coord_t& mn, const coord_t& mx);
	void setIndex();
	void initVertex();
	void getHeightRange(float &low, float &high) const;
	coord_t BGround::getCoord(const ist::vector2& pos) const;
	ist::vector4 BGround::getRayCross(const ist::vector4 &eye, const ist::vector4 &ray) const;
	vertex_t* operator[](int i) { return &vertex[i*x_div]; }
	const vertex_t* operator[](int i) const { return &vertex[i*x_div]; }


protected:
	vertex_cont vertex;
	index_cont index;
	ist::vector2 min_edge;
	ist::vector2 max_edge;
	float grid_size;
	int x_div;
	int z_div;
};

