#ifdef USE_PNG
#include <png.h>
#endif


#include "bitmap.h"

using namespace std;
typedef unsigned char u_char;
typedef unsigned int u_int;


namespace ist {


bool Bitmap::load(const std::string& filename)
{
	clear();

	if(    strncmp(&filename[filename.size()-3], "tga", 3)==0 && loadTGA(filename)
		|| strncmp(&filename[filename.size()-3], "bmp", 3)==0 && loadBMP(filename)
		|| strncmp(&filename[filename.size()-3], "png", 3)==0 && loadPNG(filename) )
		return true;
	else
		return false;
}

bool Bitmap::save(const std::string& filename) const
{
	if(    strncmp(&filename[filename.size()-3], "tga", 3)==0 && saveTGA(filename)
		|| strncmp(&filename[filename.size()-3], "bmp", 3)==0 && saveBMP(filename)
		|| strncmp(&filename[filename.size()-3], "png", 3)==0 && savePNG(filename) )
		return true;
	else
		return false;
}




struct BMPHEAD {
	char B;
	char M;
	int file_size;
	short reserve1;
	short reserve2;
	int offset;

	BMPHEAD() {
		memset(this, 0, sizeof(*this));
		B='B';
		M='M';
		offset=54;
	}
};

struct BMPINFOHEAD {
	int header_size;
	int width;
	int height;
	short plane;
	short bits;
	int compression;
	int comp_image_size;
	int x_resolution;
	int y_resolution;
	int pallete_num;
	int important_pallete_num;

	BMPINFOHEAD() {
		memset(this, 0, sizeof(*this));
		header_size=sizeof(*this);
		plane=1;
		bits=24;
	}
};

struct TGAHEAD {
	u_char No_ID;
	u_char CMap_Type;
	u_char image_type;
	u_char CMap_Spec[5];
	short Ox;
	short Oy;
	short width;
	short height;
	u_char pixel;
	u_char IDesc_Type;

	TGAHEAD() {
		memset(this, 0, sizeof(*this));
		pixel=32;
		IDesc_Type=8;
	}
};


static bRGBA Read1Pixel(FILE *&read)
{
	bRGBA t;
	fread(&t.b, sizeof(u_char), 1, read);
	fread(&t.g, sizeof(u_char), 1, read);
	fread(&t.r, sizeof(u_char), 1, read);
	fread(&t.a, sizeof(u_char), 1, read);
	return t;
}


namespace RLECompress {
	vector<u_char> comp_pixel;


	static void WriteSameData(vector<bRGBA> &temp_pixel)
	{
		comp_pixel.push_back( temp_pixel.size()+0x80 );

		comp_pixel.push_back( temp_pixel[0].b );
		comp_pixel.push_back( temp_pixel[0].g );
		comp_pixel.push_back( temp_pixel[0].r );
		comp_pixel.push_back( temp_pixel[0].a );

		temp_pixel.clear();
	}

	static void WriteDifferentData(vector<bRGBA> &temp_pixel)
	{
		comp_pixel.push_back( temp_pixel.size()-1 );

		for(u_int d=0; d<temp_pixel.size(); d++) {
			comp_pixel.push_back( temp_pixel[d].b );
			comp_pixel.push_back( temp_pixel[d].g );
			comp_pixel.push_back( temp_pixel[d].r );
			comp_pixel.push_back( temp_pixel[d].a );
		}

		temp_pixel.clear();
	}

	static void Compress(const bRGBA *start, int width)
	{
		vector<bRGBA> same, diff;

		for(int i=0; i!=width; ++i, ++start) {
			const bRGBA *ip=start; ++ip;
			bRGBA dist=*start;

			if( i+1!=width && dist==*ip && same.size()<0x79 ) {
				same.push_back(dist);

				if(diff.size()!=0)
					WriteDifferentData(diff);
			}
			else  {
				if(same.size()>0x00) {
					WriteSameData(same);
				}
				else {
					diff.push_back(dist);

					if(diff.size()==0x79 )
						WriteDifferentData(diff);
				}
			}
		}

		if(same.size()!=0x00)
			WriteSameData(same);
		else if(diff.size()!=0)
			WriteDifferentData(diff);
	}
}



// BMP

bool Bitmap::loadBMP(const std::string& filename)
{
	u_int j;
	FILE *read;
	BMPHEAD head;
	BMPINFOHEAD infohead;

	read=fopen(filename.c_str(), "rb");
	if(read==0)
		return false;

	fread(&head.B,         sizeof(char), 1, read);
	fread(&head.M,         sizeof(char), 1, read);
	fread(&head.file_size, sizeof(int) , 1, read);
	fread(&head.reserve1,  sizeof(short),1, read);
	fread(&head.reserve2,  sizeof(short),1, read);
	fread(&head.offset,    sizeof(int),  1, read);

	fread(&infohead.header_size,sizeof(int), 1, read);
	fread(&infohead.width,  sizeof(int), 1, read);
	fread(&infohead.height, sizeof(int), 1, read);
	fread(&infohead.plane,  sizeof(short), 1, read);
	fread(&infohead.bits,   sizeof(short), 1, read);
	fread(&infohead.compression,    sizeof(int), 1, read);
	fread(&infohead.comp_image_size,sizeof(int), 1, read);
	fread(&infohead.x_resolution,   sizeof(int), 1, read);
	fread(&infohead.y_resolution,   sizeof(int), 1, read);
	fread(&infohead.pallete_num,    sizeof(int), 1, read);
	fread(&infohead.important_pallete_num, sizeof(int), 1, read);

	if(head.B!='B' || head.M!='M' || infohead.bits!=24 )
		return false;

	resize(infohead.width, infohead.height);

	for(int i=getHeight()-1; i>=0; --i) {
		for(int j=0; j<getWidth(); ++j) {
			bRGBA& c = (*this)[i][j];

			fread(&c.b, sizeof(u_char), 1, read);
			fread(&c.g, sizeof(u_char), 1, read);
			fread(&c.r, sizeof(u_char), 1, read);
			c.a=(*this)[i][j].r;
		//	c.a=255;
		}
	}
	fclose(read);

	return true;
}


bool Bitmap::saveBMP(const std::string& filename) const
{
	if(empty()) return false;

	FILE *write = fopen(filename.c_str(), "wb");
	if(write==0)
		return false;
	BMPHEAD head;
	BMPINFOHEAD infohead;

	head.file_size = sizeof(BMPHEAD)+sizeof(BMPINFOHEAD)+getWidth()*getHeight()*3;
	infohead.width = getWidth();
	infohead.height = getHeight();

	fwrite(&head.B,         sizeof(char), 1, write);
	fwrite(&head.M,         sizeof(char), 1, write);
	fwrite(&head.file_size, sizeof(int) , 1, write);
	fwrite(&head.reserve1,  sizeof(short),1, write);
	fwrite(&head.reserve2,  sizeof(short),1, write);
	fwrite(&head.offset,    sizeof(int),  1, write);

	fwrite(&infohead.header_size,sizeof(int), 1, write);
	fwrite(&infohead.width,  sizeof(int), 1, write);
	fwrite(&infohead.height, sizeof(int), 1, write);
	fwrite(&infohead.plane,  sizeof(short), 1, write);
	fwrite(&infohead.bits,   sizeof(short), 1, write);
	fwrite(&infohead.compression,    sizeof(int), 1, write);
	fwrite(&infohead.comp_image_size,sizeof(int), 1, write);
	fwrite(&infohead.x_resolution,   sizeof(int), 1, write);
	fwrite(&infohead.y_resolution,   sizeof(int), 1, write);
	fwrite(&infohead.pallete_num,    sizeof(int), 1, write);
	fwrite(&infohead.important_pallete_num, sizeof(int), 1, write);

	for(int i=getHeight()-1; i>=0; --i) {
		for(int j=0; j<getWidth(); ++j) {
			const bRGBA& c = (*this)[i][j];
			fwrite(&c.b, sizeof(u_char), 1, write);
			fwrite(&c.g, sizeof(u_char), 1, write);
			fwrite(&c.r, sizeof(u_char), 1, write);
		}
	}
	fclose(write);

	return true;
}





// TGA

bool Bitmap::loadTGA(const std::string& filename)
{
	FILE *read = fopen(filename.c_str(), "rb");
	if(read==0) return false;

	TGAHEAD head;
	fread(&head.No_ID,     sizeof(u_char), 1, read);
	fread(&head.CMap_Type, sizeof(u_char), 1, read);
	fread(&head.image_type,sizeof(u_char), 1, read);
	fread(&head.CMap_Spec, sizeof(u_char), 5, read);
	fread(&head.Ox,        sizeof(short), 1, read);
	fread(&head.Oy,        sizeof(short), 1, read);
	fread(&head.width,     sizeof(short), 1, read);
	fread(&head.height,    sizeof(short), 1, read);
	fread(&head.pixel,     sizeof(u_char), 1, read);
	fread(&head.IDesc_Type,sizeof(u_char), 1, read);

	if(head.pixel!=32)
		return false;

	resize(head.width, head.height);

	for(int i=getHeight()-1; i>=0; --i) {
		if(head.image_type==2) {
			for(int j=0; j<getWidth(); j++)
				(*this)[i][j]=Read1Pixel(read);
		}
		else if(head.image_type==10) {
			int loaded = 0;

			while(loaded<getWidth()) {
				u_char dist;

				fread(&dist, sizeof(u_char), 1, read);
				if( dist<0x80) {
					for(int j=0; j<dist+1; ++j, ++loaded)
						(*this)[i][loaded]=Read1Pixel(read);
				}
				else {
					bRGBA t=Read1Pixel(read);
					for(int j=0x80; j<dist+1; ++j, ++loaded)
						(*this)[i][loaded]=t;
				}
			}
		}
	}
	fclose(read);

	return true;
}


bool Bitmap::saveTGA(const std::string& filename) const
{
	if(empty())
		return false;

	TGAHEAD head;
	head.width = getWidth();
	head.height = getHeight();
	head.image_type = 10;

	FILE *write = fopen(filename.c_str(), "wb");
	if(write==0)
		return false;
	fwrite(&head.No_ID,     sizeof(u_char), 1, write);
	fwrite(&head.CMap_Type, sizeof(u_char), 1, write);
	fwrite(&head.image_type,sizeof(u_char), 1, write);
	fwrite(&head.CMap_Spec, sizeof(u_char), 5, write);
	fwrite(&head.Ox,        sizeof(short), 1, write);
	fwrite(&head.Oy,        sizeof(short), 1, write);
	fwrite(&head.width,     sizeof(short), 1, write);
	fwrite(&head.height,    sizeof(short), 1, write);
	fwrite(&head.pixel,     sizeof(u_char), 1, write);
	fwrite(&head.IDesc_Type,sizeof(u_char), 1, write);

	for(int i=getHeight()-1; i>=0; --i)
		RLECompress::Compress((*this)[i], getWidth());

	for(vector<u_char>::iterator p=RLECompress::comp_pixel.begin(); p!=RLECompress::comp_pixel.end(); ++p)
			fwrite(&(*p), sizeof(u_char), 1, write);

	fclose(write);

	vector<u_char>().swap(RLECompress::comp_pixel);

	return true;
}



// PNG
#ifdef USE_PNG

bool Bitmap::loadPNG(const std::string& filename) {
	FILE *fp = fopen(filename.c_str(), "rb");
	if(fp==0)
		return false;

	png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	if(png_ptr==0) {
		fclose(fp);
		return false;
	}

	png_infop info_ptr = png_create_info_struct(png_ptr);
	if(info_ptr==0) {
		fclose(fp);
		png_destroy_read_struct(&png_ptr, png_infopp_NULL, png_infopp_NULL);
		return false;
	}

	png_init_io(png_ptr, fp);

	png_uint_32 w, h;
	int bit_depth, color_type, interlace_type;

	png_read_info(png_ptr, info_ptr);
	png_get_IHDR(png_ptr, info_ptr, &w, &h, &bit_depth, &color_type, &interlace_type, int_p_NULL, int_p_NULL);

	resize(w, h);

	png_set_strip_16(png_ptr);
	png_set_packing(png_ptr);
	if(color_type==PNG_COLOR_TYPE_PALETTE)
		png_set_palette_to_rgb(png_ptr);
	if(color_type == PNG_COLOR_TYPE_GRAY && bit_depth<8)
		png_set_gray_1_2_4_to_8(png_ptr);
	if(png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
		png_set_tRNS_to_alpha(png_ptr);
	png_read_update_info(png_ptr, info_ptr);


	std::vector<png_bytep> row_pointers(getHeight());
	for(int row=0; row<getHeight(); ++row)
		row_pointers[row] = (png_bytep)png_malloc(png_ptr, png_get_rowbytes(png_ptr, info_ptr));
	png_read_image(png_ptr, &row_pointers[0]);

	if(color_type==PNG_COLOR_TYPE_RGB_ALPHA) {
		for(int i=0; i<getHeight(); ++i) {
			for(int j=0; j<getWidth(); ++j) {
				bRGBA& c = (*this)[i][j];
				c.r = row_pointers[i][j*4+0];
				c.g = row_pointers[i][j*4+1];
				c.b = row_pointers[i][j*4+2];
				c.a = row_pointers[i][j*4+3];
			}
		}
	}
	else if(color_type==PNG_COLOR_TYPE_RGB) {
		for(int i=0; i<getHeight(); ++i) {
			for(int j=0; j<getWidth(); ++j) {
				bRGBA& c = (*this)[i][j];
				c.r = row_pointers[i][j*3+0];
				c.g = row_pointers[i][j*3+1];
				c.b = row_pointers[i][j*3+2];
				c.a = 255;
			}
		}
	}

	for(int row=0; row<getHeight(); ++row)
		png_free(png_ptr, row_pointers[row]);

	png_read_end(png_ptr, info_ptr);
	png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
	fclose(fp);
	return true;
}

bool Bitmap::savePNG(const std::string& filename) const
{
	FILE *fp = fopen(filename.c_str(), "wb");
	if(fp==0)
	return false;

	png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);

	if(png_ptr==0) {
		fclose(fp);
		return false;
	}

	png_infop info_ptr = png_create_info_struct(png_ptr);
	if(info_ptr==0) {
		fclose(fp);
		png_destroy_write_struct(&png_ptr,  png_infopp_NULL);
		return false;
	}

	png_init_io(png_ptr, fp);
	png_set_IHDR(png_ptr, info_ptr, getWidth(), getHeight(), 8,
		PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
	png_write_info(png_ptr, info_ptr);

	std::vector<bRGBA> tmp = data;
	std::vector<png_bytep> row_pointers(getHeight());
	for(int i=0; i<getHeight(); ++i) {
		row_pointers[i] = tmp[i*getWidth()].v;
	}

	png_write_image(png_ptr, &row_pointers[0]);

	png_write_end(png_ptr, info_ptr);
	png_destroy_write_struct(&png_ptr, &info_ptr);
	fclose(fp);
	return true;
}

#else

bool Bitmap::loadPNG(const std::string& filename) { return false; }
bool Bitmap::savePNG(const std::string& filename) const { return false; }

#endif


}	// ist


