#include <stdexcept>
#include "collision_detector.h"

using namespace std;


namespace ist {
namespace i3d {



ICollisionModel::~ICollisionModel() {}



CollisionDetector::CollisionDetector() :
	result(false), distance(0)
{}

//|S-|S
template <>
bool CollisionDetector::detect_<IPolygonCollisionModel, IPolygonCollisionModel>
	(const IPolygonCollisionModel& base, const IPolygonCollisionModel& target)
{
	throw std::runtime_error("not implemented");

/*	if(base.chunc==0 || target.chunc==0
	|| !base.bb.overlap(target.bb) )
		return false;

	const BasePolygonChunc::line_cont &line_array = base.chunc->line_array;
	for(BasePolygonChunc::line_cont::const_iterator p=line_array.begin(); p!=line_array.end(); ++p) {
		const vector4& head = base.t_vertex[ p->getIndex(0) ];
		const vector4& tail = base.t_vertex[ p->getIndex(1) ];

		for(unsigned int i=0; i<target.chunc->getFaceNum(); ++i) {
			const vector4& normal = target.t_normal[i];
			const vector4 *vtx[3] = {
				&target.t_vertex[ target.chunc->idx_array[i*3+0] ],
				&target.t_vertex[ target.chunc->idx_array[i*3+1] ],
				&target.t_vertex[ target.chunc->idx_array[i*3+2] ],
			};
		
			float dist_tri = normal.dot( *vtx[0] );
			float dist_tail= normal.dot(tail);
			if( (normal.dot( head )-dist_tri)*(dist_tail-dist_tri)>0)
				continue;

			vector4 direction=(head-tail).normalize();
			float t=-(normal.dot(tail)-normal.dot(*vtx[0]))/normal.dot(direction);
			vector4 proj_pos=tail+direction*t;

			float inside=0;
			for(unsigned int j=0; j<3; ++j)
				inside+=(*vtx[j]-proj_pos).cos( *vtx[(j+1)%3]-proj_pos );

			if( inside>358.0f*radian ) {
				distance = base.chunc->bounding_sphere.radius;
				collide_pos=proj_pos;
				collide_normal=normal;
				return true;
			}
		}
	}
	
	return false;
*/
}

//|S-_
template <>
bool CollisionDetector::detect_<IPolygonCollisionModel, IPointCollisionModel>
	(const IPolygonCollisionModel& base, const IPointCollisionModel& target)
{
	const IPolygonCollisionModel::vertex_cont &b_vtx = base.getVertex();
	const IPolygonCollisionModel::normal_cont &b_nml = base.getNormal();
	const IPolygonCollisionModel::index_cont  &b_idx = base.getIndex();
	const vector4& t_pos = target.getPos();
	float t_radius = target.getRadius();

	if(!base.getBoundingBox().inner_point(t_pos, t_radius))
		return false;


	bool is_inside = false;
	float min_distance = 0;
	vector4 last_collide_pos;
	vector4 last_collide_normal;

	for(unsigned int i=0; i<(b_idx.size()/3); ++i) {
		const int *triangle_index = &b_idx[i*3];
		const vector4& normal = b_nml[i];
		const vector4 *vtx[3] = {
			&b_vtx[ triangle_index[0] ],
			&b_vtx[ triangle_index[1] ],
			&b_vtx[ triangle_index[2] ],
		};

		for(int i=0; i<3; ++i) { // _ƏՓ˂Ă邩
			float d = (*vtx[i]-t_pos).norm();
			if(d<t_radius) {
				distance = d;
				collide_pos = *vtx[i];
				collide_normal = normal;
				return true;
			}
		}

		float dist1 = normal.dot( t_pos );	//_-_ԋ
		float dist2 = normal.dot( *vtx[0] );		//_-ʊԋ
		float dist3 = dist1-dist2;				//-_ԋAꂪʂ̔Αɂ

		if( fabsf(dist3)<fabsf(min_distance) || !is_inside ) {	//_Ɉԋ߂ʂT
			vector4 proj_pos = t_pos-normal*dist3;			//ʂɓeꂽ_̈ʒu

			float inside = 0;
			for(unsigned int j=0; j<3; ++j)		//Op`Ƃ̓OBe_Ɛp360Ȃɂ͂B
				inside+=(*vtx[j]-proj_pos).cos( *vtx[(j+1)%3]-proj_pos );

			if(inside>358.0f*radian) {	//2̌덷F߂
				is_inside = true;
				min_distance = dist3;

				last_collide_pos = proj_pos;
				last_collide_normal = normal;
			}
		}
	}

	if(min_distance<t_radius && is_inside) {
		distance = t_radius;
		collide_pos = last_collide_pos;
		collide_normal = last_collide_normal;
		return true;
	}
	else
		return false;
}

//_-|S
template <>
bool CollisionDetector::detect_<IPointCollisionModel, IPolygonCollisionModel>
	(const IPointCollisionModel& base, const IPolygonCollisionModel& target)
{
	return detect_(target, base);
}

//|S-
template<>
bool CollisionDetector::detect_<IPolygonCollisionModel, ILineCollisionModel>
	(const IPolygonCollisionModel& base, const ILineCollisionModel& target)
{
	if( !base.getBoundingBox().overlap(target.getBoundingBox()) )
		return false;

	const IPolygonCollisionModel::vertex_cont &b_vtx = base.getVertex();
	const IPolygonCollisionModel::normal_cont &b_nml = base.getNormal();
	const IPolygonCollisionModel::index_cont  &b_idx = base.getIndex();
	const vector4& t_head = target.getHead();
	const vector4& t_tail = target.getTail();

	bool is_inside = false;
	float min_distance = 0;
	vector4 last_collide_pos;
	vector4 last_collide_normal;

	for(unsigned int i=0; i<(b_idx.size()/3); ++i) {
		const vector4& normal = b_nml[i];
		const vector4 *vtx[3] = {
			&b_vtx[ b_idx[i*3+0] ],
			&b_vtx[ b_idx[i*3+1] ],
			&b_vtx[ b_idx[i*3+2] ],
		};
	
		float dist_tri  = normal.dot(*vtx[0]);
		float dist_tail = normal.dot(t_tail);
		if( (normal.dot(t_head)-dist_tri)*(dist_tail-dist_tri)>0)
			continue;

		vector4 direction = (t_head-t_tail).normalize();
		float t = -(normal.dot(t_tail)-normal.dot(*vtx[0]))/normal.dot(direction);
		vector4 proj_pos = t_tail+direction*t;

		float inside=0;
		for(unsigned int j=0; j<3; ++j) {
			if((proj_pos-*vtx[j]).square()<0.01) {
				inside=360.0f*radian;
				break;
			}
			inside+=(*vtx[j]-proj_pos).cos( *vtx[(j+1)%3]-proj_pos );
		}

		if( (inside>358.0f*radian)
		&&  (fabsf(dist_tail)<fabsf(min_distance) || !is_inside) ) {
			is_inside = true;
			min_distance = dist_tail;
			last_collide_pos = proj_pos;
			last_collide_normal = normal;
		}
	}

	if(is_inside) {
		distance = target.getRadius()+((t_tail-last_collide_pos).norm()+min_distance);
		collide_pos = last_collide_pos;
		collide_normal = last_collide_normal;
		return true;
	}
	else
		return false;
}

//-|S
template<>
bool CollisionDetector::detect_<ILineCollisionModel, IPolygonCollisionModel>
	(const ILineCollisionModel& base, const IPolygonCollisionModel& target)
{
	return detect_(target, base);
}



//_-_
template <>
bool CollisionDetector::detect_<IPointCollisionModel, IPointCollisionModel>
	(const IPointCollisionModel& base, const IPointCollisionModel& target)
{
	if( !base.getBoundingBox().overlap(target.getBoundingBox()) )
		return false;

	const vector4 &b_pos = base.getPos();
	const vector4 &t_pos = target.getPos();

	vector4 pos_gap = t_pos-b_pos;
	float union_radius = base.getRadius() + target.getRadius();

	if(pos_gap.norm() < union_radius) {
		distance = union_radius;
		collide_normal = vector4(pos_gap).normalize();
		collide_pos = b_pos;
		return true;
	}

	return false;
}


//-_
template <>
bool CollisionDetector::detect_<ILineCollisionModel, IPointCollisionModel>
	(const ILineCollisionModel& base, const IPointCollisionModel& target)
{
	if( !base.getBoundingBox().overlap(target.getBoundingBox()) )
		return false;

	const vector4 &b_head = base.getHead();
	const vector4 &b_tail = base.getTail();
	float b_radius = base.getRadius();
	const vector4 &t_pos = target.getPos();
	float t_radius = target.getRadius();

	bool _result = false;
	vector4 segment = b_head-b_tail;
	float union_radius = b_radius+t_radius;

	float u = 0;
	if(b_tail!=b_head)
		u = (t_pos-b_tail).dot(segment) / segment.square();

	if(u>0 && u<1) {
		vector4 nearest = b_tail+(segment*u);
		float distance = (t_pos-nearest).norm();

		if( distance < union_radius ) {
			collide_normal = (t_pos-nearest).normalize();
			collide_pos = nearest;
			_result=true;
		}
	}
	else if((t_pos-b_head).norm()<union_radius) {
		collide_normal = (t_pos-b_head).normalize();
		collide_pos = b_head;
		_result=true;
	}
	else if((t_pos-b_tail).norm()<union_radius) {
		collide_normal = (t_pos-b_tail).normalize();
		collide_pos = b_tail;
		_result=true;
	}

	if(_result)
		distance = union_radius;

	return _result;
}


//_-
template <>
bool CollisionDetector::detect_<IPointCollisionModel, ILineCollisionModel>
	(const IPointCollisionModel& base, const ILineCollisionModel& target)
{
	return detect_(target, base);
}


//-
template <>
bool CollisionDetector::detect_<ILineCollisionModel, ILineCollisionModel>
	(const ILineCollisionModel& base, const ILineCollisionModel& target)
{
	if( !base.getBoundingBox().overlap(target.getBoundingBox()) )
		return false;

	return false;
}


//_-2D|
template <>
bool CollisionDetector::detect_<IPointCollisionModel, IPolygon2DCollisionModel>
	(const IPointCollisionModel& base, const IPolygon2DCollisionModel& target)
{
	vector2 point_pos(base.getPos().x, base.getPos().z);
	float ang = 0;
	for(size_t i=0; i<target.getNumVertex(); ++i) {
		ang+=(target.getVertex(i)-point_pos).cos(target.getVertex((i+1)%target.getNumVertex())-point_pos);
	}
	if(ang>358.0f*radian) {
		return true;
	}

	return false;
}

//2D|-_
template <>
bool CollisionDetector::detect_<IPolygon2DCollisionModel, IPointCollisionModel>
	(const IPolygon2DCollisionModel& base, const IPointCollisionModel& target)
{
	return detect_(target, base);
}

//2D|-2D|
template <>
bool CollisionDetector::detect_<IPolygon2DCollisionModel, IPolygon2DCollisionModel>
	(const IPolygon2DCollisionModel& base, const IPolygon2DCollisionModel& target)
{
	return false;
}





//NX^
bool CollisionDetector::detect(const ICollisionModel &base, const ICollisionModel &target)
{
	result = false;

	if(base.getType()==CM_TRUE || target.getType()==CM_TRUE) {
		result = true;
	}
	else if( base.getType()==CM_POINT ) {
		if     ( target.getType()==CM_POINT )
			result=detect_( (static_cast<const IPointCollisionModel&>(base)),
							(static_cast<const IPointCollisionModel&>(target)) );
		else if( target.getType()==CM_LINE )
			result=detect_( (static_cast<const IPointCollisionModel&>(base)),
							(static_cast<const ILineCollisionModel&>(target)) );
		else if( target.getType()==CM_POLYGON )
			result=detect_( (static_cast<const IPointCollisionModel&>(base)),
							(static_cast<const IPolygonCollisionModel&>(target)) );
		else if( target.getType()==CM_POLYGON2D )
			result=detect_( (static_cast<const IPointCollisionModel&>(base)),
							(static_cast<const IPolygon2DCollisionModel&>(target)) );
	}
	else if( base.getType()==CM_LINE ) {
		if     ( target.getType()==CM_POINT )
			result=detect_( (static_cast<const ILineCollisionModel&>(base)),
							(static_cast<const IPointCollisionModel&>(target)) );
		else if( target.getType()==CM_LINE )
			result=detect_( (static_cast<const ILineCollisionModel&>(base)),
							(static_cast<const ILineCollisionModel&>(target)) );
		else if( target.getType()==CM_POLYGON )
			result=detect_( (static_cast<const ILineCollisionModel&>(base)),
							(static_cast<const IPolygonCollisionModel&>(target)) );
	}
	else if( base.getType()==CM_POLYGON ) {
		if( target.getType()==CM_POINT )
			result=detect_( (static_cast<const IPolygonCollisionModel&>(base)),
							(static_cast<const IPointCollisionModel&>(target)) );
		else if( target.getType()==CM_LINE )
			result=detect_( (static_cast<const IPolygonCollisionModel&>(base)),
							(static_cast<const ILineCollisionModel&>(target)) );
		else if( target.getType()==CM_POLYGON )
			result=detect_( (static_cast<const IPolygonCollisionModel&>(base)),
							(static_cast<const IPolygonCollisionModel&>(target)) );
	}
	else if( base.getType()==CM_POLYGON2D ) {
		if( target.getType()==CM_POINT )
			result=detect_( (static_cast<const IPolygon2DCollisionModel&>(base)),
							(static_cast<const IPointCollisionModel&>(target)) );
		else if( target.getType()==CM_POLYGON2D )
			result=detect_( (static_cast<const IPolygon2DCollisionModel&>(base)),
							(static_cast<const IPolygon2DCollisionModel&>(target)) );
	}
	return result;
}



} //i3d
} //ist

