#include <algorithm>
#include "gl_animated_polygon.h"
using namespace std;

namespace ist {
namespace i3d {


void AnimatedModel::draw(iterator p, float t) const
{
	if(p==vtx_clip.end() || p->second.empty())
		return;

	static vertex_cont t_vertex;
	static vertex_cont vertex_array;
	p->second.morph(t, t_vertex);

	vertex_array.resize(idx_array.size()*3);
	for(size_t i=0; i<idx_array.size(); ++i) {
		vertex_array[i] = t_vertex[idx_array[i]];
	}

	glVertexPointer(3, GL_FLOAT, 0, vertex_array.front().v);
	glTexCoordPointer(2, GL_FLOAT, 0, tex_array.front().v);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glDrawArrays(GL_TRIANGLES, 0, idx_array.size());
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
}



void CachedAnimatedModel::draw(iterator p, int t) const
{
	if(p==vtx_clip.end())
		return;

	const vertex_sequence& seq = p->second;
	if(seq.vertex.empty())
		return;

	const vertex_cont *vertex_array;
	if(t<=seq.beginFrame())
		vertex_array = &seq.vertex.front();
	else if(t>=seq.endFrame())
		vertex_array = &seq.vertex.back();
	else
		vertex_array = &seq.vertex[t-seq.beginFrame()];

	if(enable_normal) {
		const normal_cont *normal_array;
		if(t<seq.beginFrame())
			normal_array = &seq.normal.front();
		else if(t>=seq.endFrame())
			normal_array = &seq.normal.back();
		else
			normal_array = &seq.normal[t-seq.beginFrame()];

		glVertexPointer(3, GL_FLOAT, 0, vertex_array->front().v);
		glNormalPointer(GL_FLOAT, 0, normal_array->front().v);
		glTexCoordPointer(2, GL_FLOAT, 0, tex_array.front().v);
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glDrawArrays(GL_TRIANGLES, 0, idx_array.size());
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);
		glDisableClientState(GL_VERTEX_ARRAY);
	}
	else {
		glVertexPointer(3, GL_FLOAT, 0, vertex_array->front().v);
		glTexCoordPointer(2, GL_FLOAT, 0, tex_array.front().v);
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glDrawArrays(GL_TRIANGLES, 0, idx_array.size());
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glDisableClientState(GL_VERTEX_ARRAY);
	}
}

IAnimatedModel* IAnimatedModel::create() { return new AnimatedModel(); }
ICachedAnimatedModel* ICachedAnimatedModel::create() { return new CachedAnimatedModel(); }



} // i3d
} // ist

