#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "gl_include.h"
#include "gl_kanji.h"
#include "i3d_matrix_saver.h"
#include "i3d_screen_matrix.h"




static void sjis2jis(unsigned char &c1, unsigned char &c2)
{
	if (c1>=0xe0) { c1 = c1-0x40; }
	if (c2>=0x80) { c2 = c2-1; }
	if (c2>=0x9e) {
		c1 = (c1-0x70) * 2;
		c2 = c2-0x7d;
	} else {
		c1 = ((c1-0x70)*2)-1;
		c2 = c2-0x1f;
	}
}

static void euc2jis(unsigned char &c1, unsigned char &c2)
{
	c1-=0x80;
	c2-=0x80;
}



namespace ist {
namespace i3d {


Kanji::Kanji(int _code)
	: code(_code)
{
	init();
}

Kanji::Kanji(const std::string& filename, int _code)
	: code(_code)
{
	init();
	load(filename);
}

void Kanji::setCode(int _code) { code=_code; }
int Kanji::getFontSize() { return font_size; }

void Kanji::init()
{
	font_size = 0;
	char_byte = 0;
	max_index = 0;
}

void Kanji::print(int x, int y, const char *_str)
{
	if(_str==0 || data.empty()) return;

	const unsigned char *str = reinterpret_cast<const unsigned char*>(_str);

	MatrixSaver mmsaver(GL_MODELVIEW_MATRIX);	// ϊsޔ
	MatrixSaver pmsaver(GL_PROJECTION_MATRIX);
	ScreenMatrixIV smatrix;		// WnXN[ɂ킹

	glRasterPos2i(x, y+font_size);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	int down = 0;

	while(*str!=0) {
		if(*str=='\n') {
			down+=font_size+1;
			++str;
			glRasterPos2i(x, y+font_size+down);
			continue;
		}
		if(isprint(*str)) {	// asciȉꍇ
			if(*str < max_index)
				glBitmap(font_size, font_size, font_size/2, 0, font_size/2, 0, getChar(*str) );
			++str;
		}
		else {	// {ꕶ̏ꍇ
			unsigned char c1=str[0];
			unsigned char c2=str[1];

			switch(code) {
				case SJIS: sjis2jis(c1,c2); break;
				case EUC:  euc2jis (c1,c2); break;
			}

			int index=(c1-0x20)*96 + c2-0x20 +0xff;
			if(index<max_index && index>=0)
				glBitmap(font_size, font_size, 0, 0, font_size, 0, getChar(index) );
			str+=2;
		}
	}
}


bool Kanji::load(const std::string& filename)
{
	data.clear();
	font_size=0;
	char_byte=0;

	FILE *in = fopen(filename.c_str(), "rb");
	if(in==0)
		return false;

	fread(&font_size, sizeof(int), 1, in);

	int n = 0;
	while(n<font_size) { n+=8; }
	int char_bit = n*font_size;
	char_byte = char_bit/8;


	unsigned char length, temp;
	while(fread(&length, 1, 1, in)!=0) {
		if(length<0x80) {
			for(int i=0; i<length+1; ++i) {
				fread(&temp, 1, 1, in);
				data.push_back(temp);
			}
		}
		else {
			fread(&temp, 1, 1, in);

			for(int i=0x80; i<length+1; ++i)
				data.push_back(temp);
		}
	}
	fclose(in);

	max_index = data.size()/char_byte;
	return true;
}

const unsigned char* Kanji::getChar(int index) { return &data[index*char_byte]; }


}	// gl
}	// ist

