#include <map>
#include "gl_include.h"
#include "i3d_light.h"


namespace ist {
namespace i3d {

namespace light {
	std::map<int, const Light*> light_list;
	int light_table[8]={GL_LIGHT0, GL_LIGHT1, GL_LIGHT2, GL_LIGHT3, GL_LIGHT4, GL_LIGHT5, GL_LIGHT6, GL_LIGHT7};

	bool assign_id(const Light *lp) {
		for(int i=0; i<8; ++i) {
			if(light_list[ light_table[i] ]==0) {
				light_list[ light_table[i] ]=lp;
				return true;
			}
		}
		return false;
	}

	bool remove_id(const Light *lp) {
		for(int i=0; i<8; ++i) {
			if(light_list[ light_table[i] ]==lp) {
				light_list[ light_table[i] ]=0;
				return true;
			}
		}
		return false;
	}

	int search_id(const Light *lp) {
		static const Light *last_lp = 0;
		static int last_id = 0;

		if(lp==last_lp)
			return last_id;
		for(int i=0; i<8; ++i) {
			if(light_list[ light_table[i] ]==lp) {
				last_lp = lp;
				last_id = light_table[i];
				return light_table[i];
			}
		}
		return 0;
	}
}


Light::Light()
	: ambient(0.2, 0.2, 0.2, 1.0),
	  diffuse(0.8, 0.8, 0.8, 1.0),
	  specular(1.0, 1.0, 1.0, 1.0)
{
	light::assign_id(this);
}
	
Light::~Light()
{
	light::remove_id(this);
}

void Light::enable() const
{
	int id = light::search_id(this);
	if(id==0)
		return;
	glEnable(id);
	glLightfv(id, GL_POSITION, position.v);
	glLightfv(id, GL_AMBIENT,  ambient.v);
	glLightfv(id, GL_DIFFUSE,  diffuse.v);
	glLightfv(id, GL_SPECULAR, specular.v);
}

void Light::disable() const
{
	int id = light::search_id(this);
	if(id==0)
		return;
	glDisable(id);
}

bool Light::isValid()  const {
	return light::search_id(this)!=0;
}



Fog::Fog() : start(0), end(100.0f) {}
Fog::~Fog() { disable(); }

void Fog::enable() const
{
	glEnable(GL_FOG);
	glFogi(GL_FOG_MODE, GL_LINEAR);
	glFogf(GL_FOG_START, start);
	glFogf(GL_FOG_END, end);
	glFogfv(GL_FOG_COLOR, color.v);
}

void Fog::disable() const
{
	glDisable(GL_FOG);
}


} //gl
} //ist

