#ifndef IST_I3D_CAMERA
#define IST_I3D_CAMERA

#include <algorithm>
#include "matrix.h"
#include "collision_detector.h"

namespace ist {
namespace i3d {


class ViewFrustum : public SimplePolygonCollisionModel
{
public:
	ViewFrustum();
	void setup(const vector4& pos_, const vector4& target4, const vector3& up, float z_near_, float z_far_, float fovy, float aspect);
	int test(const vector4& center, float r) const;
	int test(const ICollisionModel& cm) const;

	vector4 pos;
	vector3 front;
	vector3 top;
	vector3 bottom;
	vector3 right;
	vector3 left;
	float z_near;
	float z_far;
};




class Camera
{
public:
	Camera();
	virtual ~Camera();
	virtual void look()=0;

	vector4 position;
	vector4 target;
	vector3 direction;
	float z_near;
	float z_far;
};


class OrthographicCamera : public Camera
{
public:
	OrthographicCamera();
	void look();
	void setScreen(float l, float r, float b, float t);

	float left;
	float right;
	float bottom;
	float top;
};

class PerspectiveCamera : public Camera
{
public:
	PerspectiveCamera();
	void look();
	void setViewFrustum();
	const ViewFrustum& getViewFrustum() const;

	ViewFrustum view_frustum;
	float fovy;
	float aspect;
};

} //gl
} //ist

#endif
