#ifndef IST_POLYGON_MODEL
#define IST_POLYGON_MODEL


#include <vector>
#include <list>
#include <string>
#include "i3d_polygon_model.h"


namespace ist {
namespace i3d {



class BPolygonChunc : public ReferenceCounter<IPolygonChunc>
{
public:
	typedef std::vector<vector3> vvertex_cont;

	BPolygonChunc();
	virtual ~BPolygonChunc();

	virtual void clear();
	virtual void draw() const;
	virtual void setup();
	virtual void setNormal();
	virtual void setBoundingVolume();

	unsigned int getFaceNum()   const { return num_face; }
	unsigned int getVertexNum() const { return vtx_array.size(); }

	const vertex_cont&   getVertex() const  { return vtx_array; }
	const normal_cont&   getNormal() const  { return nml_array; }
	const vnormal_cont&  getVNormal() const { return vnml_array; }
	const texcoord_cont& getTexcoord() const{ return tex_array; }
	const index_cont&    getIndex() const   { return idx_array; }
	const BoundingSphere& getBSphere() const{ return bsphere; }
	const BoundingBox&    getBBox() const   { return bbox; }

	vertex_cont   vtx_array;
	vvertex_cont  vvtx_array;
	normal_cont   nml_array; // ʖ̖@
	vnormal_cont  vnml_array; // _̖@
	texcoord_cont tex_array;
	index_cont    idx_array;
	BoundingBox    bbox;
	BoundingSphere bsphere;
	size_t num_face;
	float facet;
};


class BPolygonModel : public IPolygonModel
{
public:
	typedef boost::intrusive_ptr<BPolygonChunc> cchunc_ptr;

	BPolygonModel() {}
	BPolygonModel(const std::string& filename) { load(filename); }

	~BPolygonModel() {}
	bool load(const std::string& filename);
	void draw() const;
	size_t getNumChunc() { return chuncs.size(); }
	chunc_ptr getChunc(int i);
	chunc_ptr getChunc(const std::string& chuncname);

	virtual void addChunc(cchunc_ptr p, const std::string& l) { chuncs.push_back(p); labels.push_back(l); }
	virtual void setup();

	virtual BPolygonChunc* createChunc() { return new BPolygonChunc(); }

private:
	typedef std::list<cchunc_ptr> chunc_cont;
	typedef std::list<std::string> label_cont;
	chunc_cont chuncs;
	label_cont labels;

protected:
	bool loadMQO(const std::string& filename);
	bool loadOBJ(const std::string& filename);
};



// TBasePolygonChuncpclass
template<class T>
class TPolygonModel : public BPolygonModel
{
public:
	typedef T chunc_type;

	TPolygonModel() {}
	TPolygonModel(const std::string& filename) { load(filename); }

	chunc_type* createChunc() { return new chunc_type(); }
};


} //i3d
} //ist


#endif


