#ifndef _RGBA_INCLUDE_
#define _RGBA_INCLUDE_


namespace ist {


template <typename T>
struct RGBA {
	union {
		struct { T r,g,b,a; };
		T v[4];
	};

	RGBA<T> operator + (RGBA<T> &t) const { return RGBA<T>(r+t.r, g+t.g, b+t.b, a+t.a ); }
	RGBA<T> operator - (RGBA<T> &t) const { return RGBA<T>(r-t.r, g-t.g, b-t.b, a-t.a ); }
	RGBA<T> operator * (RGBA<T> &t) const { return RGBA<T>(r*t.r, g*t.g, b*t.b, a*t.a ); }
	RGBA<T> operator / (RGBA<T> &t) const { return RGBA<T>(r/t.r, g/t.g, b/t.b, a/t.a ); }
	RGBA<T>& operator +=(RGBA<T> &t) { *this=*this+t; return *this; }
	RGBA<T>& operator -=(RGBA<T> &t) { *this=*this-t; return *this; }
	RGBA<T>& operator *=(RGBA<T> &t) { *this=*this*t; return *this; }
	RGBA<T>& operator /=(RGBA<T> &t) { *this=*this/t; return *this; }
	bool operator ==(const RGBA<T> &t) const { return (r==t.r && g==t.g && b==t.b && a==t.a); }
	bool operator !=(const RGBA<T> &t) const { return !(*this==t); }
	T& operator [](int i) { return v[i]; }
	const T& operator [](int i) const { return v[i]; }

	RGBA<T>() : r(0), g(0), b(0), a(0) {}
	RGBA<T>(T _r, T _g, T _b, T _a) : r(_r), g(_g), b(_b), a(_a) {}

	T shin() const { return (r*0.299f + g*0.587f + b*0.114f); }
};

typedef RGBA<unsigned char> bRGBA;
typedef RGBA<float> fRGBA;

inline fRGBA bRGBA_to_fRGBA(const bRGBA& b) { return fRGBA(float(b.r)/255.0f, float(b.g)/255.0f, float(b.b)/255.0f, float(b.a)/255.0f ); }


}	// ist

#endif
